
// $Header$
// Copyright (C) 1995-2001 by Digital Mars
// All Rights Reserved
// www.digitalmars.com
// Written by Walter Bright

#if __SC__ || __RCC__
#pragma once
#endif

#ifndef __FLTPNT_H
#define __FLTPNT_H 1

#if __cplusplus
extern "C" {
#endif

typedef float float_t;
typedef double double_t;

#if __SC__ >= 0x300
#define INFINITY	__inf
#define	NAN		__nan
#define NANS		__nans
#endif

#define	FP_NANS		0
#define FP_NANQ		1
#define	FP_INFINITE	2
#define	FP_NORMAL	3
#define	FP_SUBNORMAL	4
#define	FP_ZERO		5

#define FP_NAN		FP_NANQ
#define FP_EMPTY	6
#define FP_UNSUPPORTED	7

unsigned __cdecl __fpclassify_f(float);
unsigned __cdecl __fpclassify_d(double);
unsigned __cdecl __fpclassify_ld(long double);

#if __cplusplus
extern "C++" {
inline int fpclassify(float fe)	{ return __fpclassify_f(fe); }
inline int fpclassify(double fe){ return __fpclassify_d(fe); }
inline int fpclassify(long double fe)
	{ return (sizeof(long double) == sizeof(double))
		? __fpclassify_d(fe)
		: __fpclassify_ld(fe);
	}

inline int signbit(float fe)	{ return ((short *)&(fe))[1] & 0x8000; }
inline int signbit(double fe)	{ return ((short *)&(fe))[3] & 0x8000; }
inline int signbit(long double fe)
	{ return (sizeof(long double) == sizeof(double))
		? ((short *)&(fe))[3] & 0x8000
		: ((short *)&(fe))[4] & 0x8000;
	}

inline int isfinite(float fe)	{ return fpclassify(fe) >= FP_NORMAL; }
inline int isfinite(double fe)	{ return fpclassify(fe) >= FP_NORMAL; }
inline int isfinite(long double fe) { return fpclassify(fe) >= FP_NORMAL; }

inline int isnormal(float fe)	{ return fpclassify(fe) == FP_NORMAL; }
inline int isnormal(double fe)	{ return fpclassify(fe) == FP_NORMAL; }
inline int isnormal(long double fe) { return fpclassify(fe) == FP_NORMAL; }

inline int isnan(float fe)	{ return fpclassify(fe) <= FP_NANQ;   }
inline int isnan(double fe)	{ return fpclassify(fe) <= FP_NANQ;   }
inline int isnan(long double fe) { return fpclassify(fe) <= FP_NANQ;   }

inline int isinf(float fe)	{ return fpclassify(fe) == FP_INFINITE; }
inline int isinf(double fe)	{ return fpclassify(fe) == FP_INFINITE; }
inline int isinf(long double fe) { return fpclassify(fe) == FP_INFINITE; }

}
#else
#define fpclassify(fe)	(sizeof(fe) == sizeof(float) ?			\
			    __fpclassify_f(fe) :			\
			    (sizeof(fe) == sizeof(double) ? __fpclassify_d(fe) : __fpclassify_ld(fe)) \
			)
#define signbit(fe)	(sizeof(fe) == sizeof(float) ?			\
			    (int)(((short *)&(fe))[1] & 0x8000) :	\
			    (sizeof(fe) == sizeof(double) ? (int)(((short *)&(fe))[3] & 0x8000)  \
							  : (int)(((short *)&(fe))[4] & 0x8000)) \
			)
#define isfinite(fe)	(fpclassify(fe) >= FP_NORMAL)
#define isnormal(fe)	(fpclassify(fe) == FP_NORMAL)
#define isnan(fe)	(fpclassify(fe) <= FP_NANQ)
#define isinf(fe)	(fpclassify(fe) == FP_INFINITE)
#endif

double		__cdecl copysign(double ,double );
float		__cdecl copysignf(float ,float );
long double	__cdecl copysignl(long double, long double);

double __cdecl logb(double);
float __cdecl logbf(float);
long double __cdecl logbl(long double);

double __cdecl nextafter(double ,double );
float __cdecl nextafterf(float ,float );
long double __cdecl nextafterl(long double ,long double );

double __cdecl scalb(double ,long int);
float __cdecl scalbf(float ,long int);
long double __cdecl scalbl(long double ,long int);

double __cdecl nan(const char *);
float __cdecl nanf(const char *);
long double __cdecl nanl(const char *);

double __cdecl nans(const char *);
float __cdecl nansf(const char *);
long double __cdecl nansl(const char *);

double __cdecl remainder(double ,double);
float __cdecl remainderf(float ,float);
long double __cdecl remainderl(long double ,long double);

double __cdecl remquo(double ,double ,int *);
float __cdecl remquof(float ,float ,int *);
long double __cdecl remquol(long double ,long double ,int *);

double __cdecl rint(double );
float __cdecl rintf(float );
long double __cdecl rintl(long double );

double __cdecl round(double );
float __cdecl roundf(float );
long double __cdecl roundl(long double );

double __cdecl nearbyint(double);
float __cdecl nearbyintf(float);
long double __cdecl nearbyintl(long double);

double __cdecl trunc(double);
float __cdecl truncf(float);
long double __cdecl truncl(long double);

long int __cdecl rndtol(long double);
long int __cdecl rndtonl(long double);

#if !_WIN32
#define copysignl	copysign
#endif

#define logbl		logb
#define nextafterl	nextafter
#define scalbl		scalb
#define nanl		nan
#define nansl		nans
#define remainderl	remainder
#define remquol		remquo
#define rintl		rint
#define roundl		round
#define nearbyintl	nearbyint
#define truncl		trunc

/* Compiler Intrinsics  */
#if __INLINE_8087 && __SC__ >= 0x720
short __cdecl _inline_rndtos(double);
long __cdecl _inline_rndtol(double);
#if __INTSIZE == 4
long long __cdecl _inline_rntdoll(double);
#endif
#define rndtol(d) _inline_rndtol(d)
#endif

#if __cplusplus
}
#endif

#endif

