//_ strtold.c
// Copyright (c) 1985-2000 by Digital Mars
// All Rights Reserved
// Written by Walter Bright


#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<float.h>
#include	<string.h>
#include	<math.h>
#include	<fltpnt.h>
#include	<fenv.h>
//#include        <scrtl.h>
#ifdef __NT__
//#include        <setlocal.h>
#endif
#if _WIN32 && __SC__
// from \sc\src\include\setlocal.h
char * __cdecl __locale_decpoint;
void __pascal __set_errno (int an_errno);
#endif

#if __DMC__ >= 0x773 && _WIN32

struct longdouble
{
	unsigned short value[5];
};

#if 0

// These tables are powers of 10, but for bootstrapping purposes, use hex constants
// instead of relying on compiler to get all the bits exactly right

static long double negtab[] =
	{1e-256,1e-128,1e-64,1e-32,1e-16,1e-8,1e-4,1e-2,1e-1,1.0};

static long double postab[] =
	{1e+256,1e+128,1e+64,1e+32,1e+16,1e+8,1e+4,1e+2,1e+1};

#elif _WIN32

// We can rely on the 8087 being available, so for extra accuracy do everything to full 80 bits

static struct longdouble negtab[] =
{
    { 0x19ca, 0x637a, 0x4325, 0xc031, 0x3cac },
    { 0xe4f4, 0x64bc, 0x467c, 0xddd0, 0x3e55 },
    { 0xe9c5, 0xa539, 0xea27, 0xa87f, 0x3f2a },
    { 0x94ce, 0x4539, 0x1ead, 0xcfb1, 0x3f94 },
    { 0xe166, 0xc44d, 0x94be, 0xe695, 0x3fc9 },
    { 0xcf01, 0x8461, 0x7711, 0xabcc, 0x3fe4 },
    { 0x652e, 0xe219, 0x1758, 0xd1b7, 0x3ff1 },
    { 0xd70b, 0x70a3, 0x0a3d, 0xa3d7, 0x3ff8 },
    { 0xcccd, 0xcccc, 0xcccc, 0xcccc, 0x3ffb },
    { 0x0000, 0x0000, 0x0000, 0x8000, 0x3fff },
};

static struct longdouble postab[] =
{
    { 0xde8c, 0x9df9, 0xebfb, 0xaa7e, 0x4351 },
    { 0x8cdf, 0x80e9, 0x47c9, 0x93ba, 0x41a8 },
    { 0xa6d5, 0xffcf, 0x1f49, 0xc278, 0x40d3 },
    { 0xb59e, 0x2b70, 0xada8, 0x9dc5, 0x4069 },
    { 0x0000, 0x0400, 0xc9bf, 0x8e1b, 0x4034 },
    { 0x0000, 0x0000, 0x2000, 0xbebc, 0x4019 },
    { 0x0000, 0x0000, 0x0000, 0x9c40, 0x400c },
    { 0x0000, 0x0000, 0x0000, 0xc800, 0x4005 },
    { 0x0000, 0x0000, 0x0000, 0xa000, 0x4002 },
};

#else

// Do things in explicit hex for accuracy
static unsigned long hnegtab[][2] =
{
    { 0x64ac6f43, 0x0ac80628 },
    { 0xcf8c979d, 0x255bba08 },
    { 0x44f4a73d, 0x32a50ffd },
    { 0xd5a8a733, 0x3949f623 },
    { 0x97d889bc, 0x3c9cd2b2 },
    { 0xe2308c3a, 0x3e45798e },
    { 0xeb1c432d, 0x3f1a36e2 },
    { 0x47ae147b, 0x3f847ae1 },
    { 0x9999999a, 0x3fb99999 },
    { 0x00000000, 0x3ff00000 },
};

static unsigned long hpostab[][2] =
{
    { 0x7f73bf3c, 0x75154fdd },
    { 0xf9301d32, 0x5a827748 },
    { 0xe93ff9f5, 0x4d384f03 },
    { 0xb5056e17, 0x4693b8b5 },
    { 0x37e08000, 0x4341c379 },
    { 0x00000000, 0x4197d784 },
    { 0x00000000, 0x40c38800 },
    { 0x00000000, 0x40590000 },
    { 0x00000000, 0x40240000 },
};

#define negtab ((double *)hnegtab)
#define postab ((double *)hpostab)
#endif

#if __ZTC__ < 0x220	// newer compiler has these special values built in

#undef NAN
static unsigned long nanarray[2] = {1,0x7FF80000 };
#define NAN	(*(double *)nanarray)

#undef NANS
static unsigned long nansarray[2] = {1,0x7FF00000 };
#define NANS	(*(double *)nansarray)

#undef INFINITY
static unsigned long infinityarray[2] = {0,0x7FF00000 };
#define INFINITY	(*(double *)infinityarray)

#endif

/*************************
 * Convert string to double.
 * Terminates on first unrecognized character.
 */

long double strtold(const char *p,char **endp)
{
	long double ldval;
	int exp;
	long long msdec,lsdec;
	unsigned long msscale;
	char hex,dot,sign,subject;
	int pow;
	const char *pinit = p;
	static char infinity[] = "infinity";
	static char nans[] = "nans";
	fenv_t flagp;
	unsigned int old_cw;
	unsigned int old_status;

	fegetenv(&flagp);  /* Store all exceptions, and current status word */
	if (_8087)
	{
	    // disable exceptions from occurring, set max precision, and round to nearest
#if 1
	    __asm
	    {
		fstcw	word ptr old_cw
		mov	EAX,old_cw
		mov	ECX,EAX
		and	EAX,0xf0c0
		or	EAX,033fh
		mov	old_cw,EAX
		fldcw	word ptr old_cw
		mov	old_cw,ECX
	    }
#else
	    old_cw = _control87(_MCW_EM | _PC_64  | _RC_NEAR,
				_MCW_EM | _MCW_PC | _MCW_RC);
#endif
	}
	
	while (isspace(*p))
	    p++;
	sign = 0;			/* indicating +			*/
	switch (*p)
	{	case '-':
			sign++;
			/* FALL-THROUGH */
		case '+':
			p++;
	}
	subject = 0;
	ldval = 0.0;
	dot = 0;			/* if decimal point has been seen */
	exp = 0;
	hex = 0;
	msdec = lsdec = 0;
	msscale = 1;

	switch (*p)
	{   case 'i':
	    case 'I':
		if (memicmp(p,infinity,8) == 0)
		{   p += 8;
		    goto L4;
		}
		if (memicmp(p,infinity,3) == 0)		/* is it "inf"?	*/
		{   p += 3;
		L4:
		    ldval = HUGE_VAL;
		    subject = 1;
		    goto L3;
		}
		break;
	    case 'n':
	    case 'N':
		if (memicmp(p,nans,4) == 0)		/* "nans"?	*/
		{   p += 4;
		    ldval = NANS;
		    goto L5;
		}
		if (memicmp(p,nans,3) == 0)		/* "nan"?	*/
		{   p += 3;
		    ldval = NAN;
		L5:
		    if (*p == '(')		/* if (n-char-sequence)	*/
			goto L1;		/* invalid input	*/
		    subject = 1;
		    goto L3;
		}
	}

	if (*p == '0' && (p[1] == 'x' || p[1] == 'X'))
	{   p += 2;
	    hex++;
	    while (1)
	    {	int i = *p;

		while (isxdigit(i))
		{
		    subject = 1;	/* must have at least 1 digit	*/
		    i = isalpha(i) ? ((i & ~0x20) - ('A' - 10)) : i - '0';
		    if (msdec < (0xFFFFFFFFFFFFLL-16)/16)
			msdec = msdec * 16 + i;
		    else if (msscale < (0xFFFFFFFFFFFFLL-16)/16)
		    {	lsdec = lsdec * 16 + i;
			msscale *= 16;
		    }
		    else
			exp++;
		    exp -= dot;
		    i = *++p;
		}
#ifdef __NT__
		if (i == *__locale_decpoint && !dot)
#else
		if (i == '.' && !dot)
#endif
		{	p++;
			dot++;
		}
		else
			break;
	    }
	    exp *= 4;
	    if (!subject)		/* if error (no digits seen)	*/
		goto L1;
	    if (*p == 'p' || *p == 'P')
		goto L2;
	    else
	    {	subject = 0;
		goto L1;		/* error, exponent is req'd	*/
	    }
	}
	else
	{
	    while (1)
	    {	int i = *p;

		while (isdigit(i))
		{
		    subject = 1;	/* must have at least 1 digit	*/
		    if (msdec < (0x7FFFFFFFFFFFLL-10)/10)
			msdec = msdec * 10 + (i - '0');
		    else if (msscale < (0x7FFFFFFFFFFFLL-10)/10)
		    {	lsdec = lsdec * 10 + (i - '0');
			msscale *= 10;
		    }
		    else
			exp++;
		    exp -= dot;
		    i = *++p;
		}
#ifdef __NT__
		if (i == *__locale_decpoint && !dot)
#else
		if (i == '.' && !dot)
#endif
		{	p++;
			dot++;
		}
		else
			break;
	    }
	}
	if (!subject)			/* if error (no digits seen)	*/
	    goto L1;			/* return 0.0			*/
	if (*p == 'e' || *p == 'E')
	{
	    char sexp;
	    int e;
	L2:
	    sexp = 0;
	    switch (*++p)
	    {	case '-':
			sexp++;
			/* FALL-THROUGH */
		case '+':
			p++;
	    }
	    subject = 0;
	    e = 0;
	    while (isdigit(*p))
	    {
		if (e < DBL_MAX_EXP*2)		/* prevent integer overflow */
		{
		    e = e * 10 + *p - '0';
		}
		p++;
		subject = 1;
	    }
	    exp += (sexp) ? -e : e;
	    if (!subject)		/* if no digits in exponent	*/
		goto L1;		/* return 0.0			*/
	}

#if _WIN32
	__asm
	{
	    fild	qword ptr msdec
	    mov		EAX,msscale
	    cmp		EAX,1
	    je		La1
	    fild	long ptr msscale
	    fmul
	    fild	qword ptr lsdec
	    fadd
	La1:
	    fstp	tbyte ptr ldval
	}
#else
	ldval = msdec;
	if (msscale != 1)		/* if stuff was accumulated in lsdec */
	    ldval = ldval * msscale + lsdec;
#endif
	if (ldval)
	{   unsigned u;

	    if (hex)
	    {
		// Exponent is power of 2, not power of 10
#if _WIN32
		__asm
		{
		    fild	word ptr exp
		    fld		tbyte ptr ldval
		    fscale				// ST(0) = ST(0) * (2**ST(1))
		    fstp	ST(1)
		    fstp	tbyte ptr ldval
		}
#else
		// ldexp() doesn't work yet for long doubles
		ldval = ldexp(ldval,exp);
#endif
		goto L6;
	    }

	    u = 0;
	    pow = 256;

#if _WIN32
	    //printf("msdec = x%x, lsdec = x%x, msscale = x%x\n",msdec,lsdec,msscale);
	    //printf("dval = %g, x%llx, exp = %d\n",dval,dval,exp);
	    __asm fld	tbyte ptr ldval
#endif

	    while (exp > 0)
	    {
		while (exp >= pow)
		{
#if _WIN32
			__asm
			{
			    mov		EAX,u
			    imul	EAX,10
			    fld		tbyte ptr postab[EAX]
			    fmul
			}
#else
			ldval *= postab[u];
#endif
			exp -= pow;
		}
		pow >>= 1;
		u++;
	    }
#if _WIN32
	    __asm fstp	tbyte ptr ldval
#endif
	    while (exp < 0)
	    {	while (exp <= -pow)
		{
#if _WIN32
			__asm
			{
			    mov		EAX,u
			    imul	EAX,10
			    fld		tbyte ptr ldval
			    fld		tbyte ptr negtab[EAX]
			    fmul
			    fstp	tbyte ptr ldval
			}
#else
			ldval *= negtab[u];
#endif
			if (ldval == 0)
				__set_errno (ERANGE);
			exp += pow;
		}
		pow >>= 1;
		u++;
	    }
#if 0
	    if (0)
	    {   int i;
		for (i = 0; i < 5; i++)
		    printf("%04x ",ldval.value[i]);
		printf("\n");
		printf("%llx\n",dval);
	    }
#endif
	L6: // if overflow occurred
	    if (ldval == HUGE_VAL)
		__set_errno (ERANGE);		// range error
	}

    L1:
	if (endp)
	{   if (subject == 0)		/* if subject string was empty	*/
		p = pinit;		/* indicate no conversion done	*/
	    *endp = (void *) p;
	}
    L3:
	fesetenv(&flagp);		// reset floating point environment
	if (_8087)
	{
	    __asm
	    {
		xor	EAX,EAX
		fstsw	AX
		fclex
		fldcw	word ptr old_cw
	    }
	}

	return (sign) ? -ldval : ldval;
}

#else

long double strtold(const char *p,char **endp)
{
    return strtod(p, endp);
}

#endif

/************************* Test ************************************/

#if 0

#include <stdio.h>
#include <float.h>
#include <errno.h>

extern "C" long double strtold(const char *p,char **endp);

struct longdouble
{
	unsigned short value[5];
};

void main()
{
    long double ld;
    struct longdouble x;
    int i;

    errno = 0;
//  ld = strtold("0x1.FFFFFFFFFFFFFFFEp16383", NULL);
    ld = strtold("0x1.FFFFFFFFFFFFFFFEp-16382", NULL);
    x = *(struct longdouble *)&ld;
    for (i = 4; i >= 0; i--)
    {
	printf("%04x ", x.value[i]);
    }
    printf("\t%d\n", errno);

    ld = strtold("1.0e5", NULL);
    x = *(struct longdouble *)&ld;
    for (i = 4; i >= 0; i--)
    {
	printf("%04x ", x.value[i]);
    }
    printf("\n");
}

#endif
